## Copyright (C) 2020 Ioannis Kosmidis

## Distributed as part of the supplementary material for the manuscript
## "Empirical bias-reducing adjustments to estimating functions"
##
## Authors: Ioannis Kosmidis
## Date: 06 January 2020
## Licence: GPL 2 or greater
## NOT A POLISHED PIECE OF PUBLIC-USE SOFTWARE!  Provided "as is".
## NO WARRANTY OF FITNESS FOR ANY PURPOSE!

## Penalized likelihood
loglik <- function(pars, X, y, link, penalize = TRUE, penalty = "det1", scores_fun, info_fun) {
    probs <- link$linkinv(drop(X %*% pars))
    loglik <- sum(dbinom(y, 1, probs, log = TRUE))
    n <- length(y)
    if (penalize) {
        j <- info_fun(pars, type = "observed")
        so <- crossprod(scores_fun(pars, contributions = TRUE))
        pen <- try(switch(penalty,
                          "det0" = 0.5 * log(det(info_fun(pars))),
                          "det1" = 0.5 * log(det(j)) - 0.5 * log(det(so)),
                          "det2" = 0.5 * n *log(det(n * diag(3) - solve(j) %*% so)),
                          "trace" = 0.5 * sum(diag(-solve(j) %*% so))), silent = TRUE)
        if (inherits(pen, "try-error")) {
            out <- NA
            return(out)
        }
        else {
            out <- loglik + pen
            attr(out, "penalty") <- pen
        }
    }
    else {
        out <- loglik
            attr(out, "penalty") <- 0
    }
    attr(out, "loglik") <- loglik
    out
}

## Assuming the the first column is ones!
## start has length ncol(X)
fit_all_models <- function(X, y, link, silent = TRUE) {
    p <- ncol(X) - 1
    inout <- list(0:1)   
    for (j in 2:p) {
        inout <- c(inout, list(0:1))
    }
    inout <- cbind(1, expand.grid(inout)) 
    n_obs <- length(y)
    n_models <- nrow(inout)    
    coefs_ml <- coefs_tr <- coefs_br <- infinite_ml <- matrix(NA, n_models, ncol(X))
    ic_names <- c("tic", "aic", "bic")
    ic_tr <- ic_ml <- ic_br <- matrix(NA, n_models, length(ic_names))
    model_names <- apply(inout, 1, function(x) paste(which(x == 1), collapse = "-"))
    separated <- converged_tr <- converged_ml <- converged_br <- logical(n_models)
    colnames(inout) <- colnames(X)
    colnames(ic_ml) <- colnames(ic_tr) <- colnames(ic_br) <- ic_names
    rownames(inout) <- rownames(ic_ml) <- rownames(ic_tr) <- rownames(ic_br) <-
        rownames(coefs_ml) <- rownames(coefs_tr) <- rownames(coefs_br) <- rownames(infinite_ml) <-
        names(separated) <- names(converged_ml) <- names(converged_br) <- names(converged_tr) <- 
        model_names
    for (j in seq.int(n_models)) {
        which_cov <- which(inout[j, ] == 1)
        X_sub <- X[, which_cov, drop = FALSE]
        ml <- glm(y ~ - 1 + X_sub, family = binomial(link), epsilon = 1e-12, maxit = 1000)
        br <- glm(y ~ - 1 + X_sub, family = binomial(link), epsilon = 1e-08, maxit = 1000,
                  method = "brglm_fit", type = "AS_mean")
        sepa <- try(glm(y ~ -1 + X_sub, family = binomial(link), method = "detect_separation"), silent = silent)
        
        if (inherits(sepa, "try-error")) {
            separated[j] <-  NA
        }
        else {
            separated[j] <- sepa$separation
            infinite_ml[j, which_cov] <- sepa$beta
        }
        
        info_fun <- get_information_function(ml)
        scores_fun <- get_score_function(ml)
        n_p <- length(which_cov)
        if (isTRUE(n_p == 1)) {
            summ <- coef(summary(ml))
            lims <- summ[, 1] + c(-1, 1) * 5 * summ[, 2]            
        }
        else {
            lims <- c(-1, 1) * Inf
        }

                
        fit_tr <- try(nlm(function(par, ...) -loglik(par, ...), p = coef(br), X = X_sub, y = y, link = link,
                      penalize = TRUE, penalty = "trace",
                      scores = scores_fun, info = info_fun), silent = silent)
        if (inherits(fit_tr, "try-error")) {
            for (m in c("BFGS", "Nelder-Mead")) {
                fit_tr <- try(optim(coef(br),
                                    fn = loglik, X = X_sub, y = y, link = link,
                                    penalize = TRUE, penalty = "trace",
                                    scores = scores_fun, info = info_fun,
                                    method = if (isTRUE(n_p == 1)) "Brent" else m,
                                    lower = lims[1],
                                    upper = lims[2],
                                    control = list(fnscale = -1, reltol = 1e-15, maxit = 1000)), silent = silent)
                if (!inherits(fit_tr, "try-error")) {
                    break
                }
            }
            scores_tr <- numDeriv::grad(loglik, fit_tr$par, X = X_sub, y = y, link = link,
                                        penalize = TRUE, penalty = "trace",
                                        scores = scores_fun, info = info_fun)
        }
        else {
            fit_tr$par <- fit_tr$estimate
            fit_tr$value <- fit_tr$minimum
            scores_tr <- fit_tr$gradient
        }
              
        scores_ml <- scores_fun(coef(ml))
        converged_tr[j] <- all(abs(scores_tr) < 1e-03)
        converged_br[j] <- all(abs(br$grad[1:n_p]) < 1e-03)
        converged_ml[j] <- all(abs(scores_ml) < 1e-03)
        
        cc_ml <- coef(ml)
        cc_tr <- fit_tr$par
        coefs_ml[j, which_cov] <- cc_ml
        coefs_br[j, which_cov] <- coef(br)
        coefs_tr[j, which_cov] <- cc_tr
        ll_ml <- loglik(coefs_ml[j, which_cov], X_sub, y, link, penalize = TRUE, penalty = "trace",
                        scores = scores_fun, info = info_fun)
        ll_tr <- loglik(coefs_tr[j, which_cov], X_sub, y, link, penalize = TRUE, penalty = "trace",
                        scores = scores_fun, info = info_fun)
        ll_br <- loglik(coefs_br[j, which_cov], X_sub, y, link, penalize = TRUE, penalty = "trace",
                        scores = scores_fun, info = info_fun)
        ic_ml[j, ] <- c(tic = -2 * ll_ml - 2 * attr(ll_ml, "penalty"),
                        aic = -2 * ll_ml + 2 * n_p,
                        bic = -2 * ll_ml + n_p * log(n_obs))
        ic_br[j, ] <- c(tic = -2 * ll_br - 2 * attr(ll_br, "penalty"),
                        aic = -2 * ll_br + 2 * n_p,
                        bic = -2 * ll_br + n_p * log(n_obs))
        ic_tr[j, ] <- c(tic = -2 * ll_tr - 2 * attr(ll_tr, "penalty"),
                        aic = -2 * ll_tr + 2 * n_p,
                        bic = -2 * ll_tr + n_p * log(n_obs))
    }    
    list(coef_ml = coefs_ml, coef_pml = coefs_tr, coef_br = coefs_br,
         ic_ml = ic_ml, ic_pml = ic_tr, ic_br = ic_br,
         separated = separated,
         converged_tr = converged_tr,
         converged_br = converged_br,
         converged_ml = converged_ml,
         infinite_ml = infinite_ml,
         models = inout)
}
