## optimization function for simulations
opt.fun <- function(theta, data, sets, n, Q, hin){
    
    ## function to put constraints on the entries of the covariance matrix
    ## to be used in function auglag (pkg alabama)
    hin <- function(x, ...)
    {
        h <- rep(NA, 3)
        h[1] <- x[1]-1e-6## sig11>0
        h[2] <- x[2]-1e-6## sig22>0
        h[3] <- x[1]*x[2]-x[3]^2-1e-6## determinant>0
        h
    }

    ## START 
    ## standard maximum pairwise likelihood estimator
    OBJ <- try(auglag(par=theta, fn=function(y) -pwlik.smith(theta=y, z=data, sets=sets, n=n, q=Q, bias.reduction=FALSE), 
                      hin=hin,
                      control=list(iter.max=500), 
                      control.outer=list(itmax=1e4, trace=FALSE, method="nlminb")), sil=TRUE)
    
    ## if optimizer made it, store results and
    ## set current estimate as a possible starting point for bias reduction
    ## note however that by default the staring point for bias reudction is theta for fair comparison
    if(class(OBJ)!="try-error")
    {
        res <- c(OBJ$pa, var.pwlik.est(z=data, theta=OBJ$pa, sets=sets, n=n, q=Q), OBJ$conv)
        start.br <- OBJ$pa
    }else{
        ## otherwise, try to get a decent starting point with SANN
        OBJ <- try(optim(pa=theta,
                         fn=function(z) pwlik.smith(data, z, sets, n, Q, bias.reduction=FALSE),
                         method="SANN", control=list(fnscale=-1, maxit=1e2)), sil=T)
        
        if(class(OBJ)!="try-error")
        {
            ## if SANN made it, use the final value as starting point
            OBJ <- try(auglag(par=OBJ$pa, fn=function(y) -pwlik.smith(theta=y, z=data, sets=sets, n=n, q=Q, bias.reduction=FALSE), 
                              hin=hin, 
                              control=list(iter.max=500),
                              control.outer=list(itmax=1e4, trace=FALSE, method="nlminb")), sil=TRUE)
            ## if optimizer succeeds, store the result
            if(class(OBJ)!="try-error"){
                res <- c(OBJ$pa, var.pwlik.est(z=data, theta=OBJ$pa, sets=sets, n=n, q=Q), OBJ$conv)			
                start.br <- OBJ$pa
            }else{
                
                res <- rep(NA, 8)				
                start.br <- theta
            }
        }else{
            res <- rep(NA, 8)				
            start.br <- theta
        }
    }
    ## ## END
    
    ## ## START 
    ## bias reduced maximum pairwise likelihood estimator
    OBJBR <- try(auglag(par=theta, fn=function(y) -pwlik.smith(theta=y, z=data, sets=sets, n=n, q=Q, bias.reduction=TRUE), 
                        hin=hin,
                        control=list(iter.max=500), 
                        control.outer=list(itmax=1e4, trace=FALSE, method="nlminb")), sil=TRUE)
    
    
    ## if optimizer made it, store results 
    if(class(OBJBR)!="try-error")
    {
        resbr <- c(OBJBR$pa, var.pwlikBR.est(z=data, theta=OBJ$pa, sets=sets, n=n, q=Q), OBJBR$conv)
    }else{
        ## otherwise try to skin the cat
        if(class(OBJ)!="try-error")
        {
            ## try a different starting point, e.g., maximum pairwise likelihood estimate at current iteration 
            OBJBR <- try(auglag(par=start.br, fn=function(y) -pwlik.smith(theta=y, z=data, sets=sets, n=n, q=Q, bias.reduction=TRUE), 
                                hin=hin, 
                                control=list(iter.max=500),
                                control.outer=list(itmax=1e4, trace=FALSE, method="nlminb")), sil=TRUE)
        }else{
            ## otherwise, try to get a decent starting point with SANN
            OBJBR <- try(optim(pa=theta,
                               fn=function(z) pwlik.smith(data, z, sets, n, Q, bias.reduction=TRUE),
                               method="SANN", control=list(fnscale=-1, maxit=1e2)), sil=T)
            ## if SANN made it, use the final value as starting point
            if(class(OBJBR)!="try-error")
                OBJBR <- try(auglag(par=OBJBR$pa, fn=function(y) -pwlik.smith(theta=y, z=data, sets=sets, n=n, q=Q, bias.reduction=TRUE), 
                                    hin=hin, 
                                    control=list(iter.max=500),	
                                    control.outer=list(itmax=1e4, trace=FALSE, method="nlminb")), sil=TRUE)
        }
	
        if(class(OBJBR)!="try-error")
            resbr <- c(OBJBR$pa, var.pwlikBR.est(z=data, theta=OBJ$pa, sets=sets, n=n, q=Q), OBJBR$conv)			
        else
            resbr <- rep(NA, 8)				
    }
    ## END

    methods <- c("ml", "trace")
    pars <- names(theta)
    data.frame(parameter = rep(pars, length(methods)),
               method = rep(methods, each = length(pars)),
               estimate = c(res[c(1:3,7)], resbr[c(1:3,7)]),
               estimate.var = c(res[4:6], 0, resbr[4:6],0),
               convergence=c(rep(res[8],4),rep(resbr[8],4)))
}





## Simulation study

## number of cores
n_cores <- 16
## set the path to save results
path <- "."

#load packages
#pairwise likelihood and trace adjustment
library(PwMev)
#the package automatically loads 
#1)"SpatialExtremes" -> data simulation
#2)"alabama" -> constraied optimization
library(parallel)

##set sample sizes and corresponding number of simulations
#base sample size
base_n <- 10
## number of base simulations
base_n_simu <- 500

multiplier <- 2^(0:4)
#sample sizes
n_obs <- base_n * multiplier
#number of simulations for each sample size
n_simu <- base_n_simu * n_obs
#number of replicates of simulations
replications <- 8
seeds <- matrix(round(runif(length(n_obs) * replications, 0, 10^6)), length(n_simu), replications)


##define configurations of dependence
#see table 1 - Padoan et al 2010
Sigma.lst <- list(Sigma1=c(sig1=300,sig2=300,sig12=0), 
                  Sigma2=c(sig1=200,sig2=300,sig12=0), 
                  Sigma3=c(sig1=200,sig2=300,sig12=150), 
                  Sigma4=c(sig1=2000,sig2=3000,sig12=1500), 
                  Sigma5=c(sig1=20,sig2=30,sig12=15))


for(SS in 1:length(Sigma.lst)){

print(SS)

#set true parameter value
which.setting <- SS
theta <- Sigma.lst[[which.setting]]


##generate sites (fixed accros simulations and replications)
set.seed(1)
#setting as in Padoan et al 2010 - section 4
n.sites <- 50
locs <- cbind(runif(n.sites, 0, 40), runif(n.sites, 0, 40))
#we'll consider all the pairs
d <- max(locs)+1
sets <- my_sets(locs, d, 1)
Q <- n.sites





##start simulation
repl <- seq.int(replications)
for (j in seq_along(n_simu)) {
    for (r in repl) {

				set.seed(seeds[j, r])
				simu.data <- matrix(
					sapply(1:n_simu[j], function(u)
						as.numeric(t(rmaxstab(n_obs[j], locs, "gauss", cov11 = theta[1], cov12 = theta[3], cov22 = theta[2], grid = FALSE)))), 		
										 ncol=n_simu[j], byrow=FALSE)

        res <- mclapply(seq.int(n_simu[j]), function(i) {
            if (i %% 100 == 0) {
                cat("n =", n_obs[j], "| replication", r, "| sample", i, "/", n_simu[j], "\n")
            }
            out <- opt.fun(theta, simu.data[,i], sets, n_obs[j], Q, hin)
            out$sample <- i
            out
        }, mc.cores = n_cores)
        
	      results <- do.call("rbind", res)
        results$true_value <- c(theta,0)
	      attr(results, "replication") <- r
	      attr(results, "N") <- n_simu[j]
	      attr(results, "n") <- n_obs[j]
	      attr(results, paste0("Sigma", which.setting)) <- 1
        save(list = "results",
             file = paste0(path, "/", names(Sigma.lst)[which.setting], "_br-mev-", j, "-repl-", r, ".rda"))
        save(list = "simu.data",
             file = paste0(path, "/DATA_", names(Sigma.lst)[which.setting], "_br-mev-", j, "-repl-", r, ".rda"))

    }
}
}
#end



#start summaries
library("dplyr")
library("ggplot2")
library(PwMev)
path="."

#LAGS <- 0:1e4
#LAGS <- cbind(LAGS, LAGS)


Sigma.lst <- list(Sigma1=c(sig1=300,sig2=300,sig12=0), 
									Sigma2=c(sig1=200,sig2=300,sig12=0), 
									Sigma3=c(sig1=200,sig2=300,sig12=150), 
									Sigma4=c(sig1=2000,sig2=3000,sig12=1500), 
									Sigma5=c(sig1=20,sig2=30,sig12=15))


which.setting <- 5
files <- dir(path, full.names = TRUE)
files <- files[grep(paste0(names(Sigma.lst)[which.setting],"_br-mev"), files)]

all_results <- NULL
for (f in files) {
    load(f)
    cres <- results
    cres$n <- attr(cres, "n")
    cres$N <- attr(cres, "N")
    cres$replication <- attr(cres, "replication")  
    all_results <- rbind(all_results, cres)
}

## Summaries conditional on the finiteness of the ML estimate
summaries <- all_results %>% group_by(n, parameter, method) %>%
	  ## exclude replications which have not converged (check the total number to do not introduce selection bias!)
	  filter(!is.na(convergence)) %>%
	  summarize(samples_perc = 100 * n() / (max(replication) * unique(N)),
	            bias = mean(estimate - true_value),
	            variance = var(estimate),
	            mc_error = sd(estimate)/sqrt(unique(N)),
	            scaled_bias = bias * unique(n),
	            scaled_bias_mc_error = mc_error * unique(n))#,
#              prob_infinite_ml = mean(infinite_ml))

pdf(paste0(path, "/",names(Sigma.lst)[which.setting],"_br-mev.pdf"))

summaries$method <- factor(summaries$method, levels = c("ml", "trace"), ordered = TRUE)
print(ggplot(summaries) +
    geom_col(aes(x = parameter, y = scaled_bias, color = method, fill = method), position = "dodge") +
    geom_hline(aes(yintercept = 0)) +
    theme_minimal() +
    ## coord_flip() +
    facet_grid( ~ n))

print(ggplot(summaries) +
    geom_col(aes(x = parameter, y = variance, color = method, fill = method), position = "dodge") +
    geom_hline(aes(yintercept = 0)) +
    theme_minimal() +
    ## coord_flip() +
    facet_grid( ~ n))

print(ggplot(summaries) +
    geom_line(aes(x = log(n, 2), y = log(abs(bias), 2), color = method)) +
    theme_minimal() +
    facet_grid(~ parameter))

dev.off()

